package cn.ischoolcode.javabook.scene9;

/* 
 * 爱校码
 * http://www.ischoolcode.cn/
 *  
 * @author zcj
 */
import java.net.*;
import java.io.*;
import java.util.*;
public class GChatServer extends Thread {
	List<GChat> chats = new LinkedList<GChat>();
	int portNum = 8888;
	boolean listening = true;
	ServerSocket serverSocket = null;

	// 从一个聊天会话接收字符串的实用方法并
	// 把它放入所有聊天对话的消息队列

	synchronized void putMsgs(String str) {
		for (int i = 0; i < chats.size(); i++) {
			GChat c =   chats.get(i);
			c.put(str);
		}
	}

	synchronized void removeMe(int index) {
		for (int i = 0; i < chats.size(); i++) {
			GChat c =  chats.get(i);
			if (index == c.myIndex) {
				c.stopChat();
				chats.remove(c);
				break;
			}
		}
		System.out.println("新的聊天大小:" + chats.size());
	}

	public void run() {
		InetAddress iaddr = null;
		Socket nextSock;
		int nThreadCount = 0;

		try {
			serverSocket = new ServerSocket(portNum);
			iaddr = InetAddress.getLocalHost();
			System.out.println("Host:" + iaddr.getHostName() + " ipaddr="
					+ iaddr.getHostAddress() + " portnum=" + portNum);

			while (listening) {
				nextSock = serverSocket.accept();
				System.out.println(nThreadCount + " 另一线程创建");

				// 创建一个线程来处理传入的连接
				GChat chatChild = new GChat(nextSock, this, nThreadCount++);
				chats.add(chatChild);
				chatChild.start();
			}

			serverSocket.close();

		} catch (IOException e) {
			System.err.println("Socket Err:" + e);
		}

	}

	public static void main(String[] args) throws IOException {
		GChatServer cserver = new GChatServer();
		if (args.length > 0) {
			cserver.portNum = Integer.parseInt(args[0]);
		}

		// 启动主服务器任务

		cserver.start();

		InputStreamReader ir = new InputStreamReader(System.in);
		BufferedReader br = new BufferedReader(ir);

		String str;
		do {
			str = br.readLine();
		} while (!"quit".equals(str));

		cserver.listening = false;
		try {
			cserver.serverSocket.close();
		} catch (IOException e) {
		}

		// 因为我们是从LinkedList删除条目，走在相反的顺序， 这进而影响后续的迭代。
		// 这需要一个图片来解释。
		for (int i = cserver.chats.size() - 1; i >= 0; i--) {
			GChat c =   cserver.chats.get(i);
			c.stopChat();
		}

	}
}

class GChat extends Thread {
	private Socket socket = null;
	int myIndex;
	boolean continueFlag = true;
	PrintWriter out = null;
	GChatServer mom;
	Deque<String> msgs = new LinkedList<String>();

	synchronized void stopChat() {
		continueFlag = false;
		notify(); // 唤醒线程等待消息
		try {
			socket.close();
		} catch (IOException e) {
		}
	}

	// 为此聊天会话添加字符串到列表
	synchronized void put(String str) {
		msgs.add(str);
		notify(); // 唤醒线程在等待这个消息
	}

	// 不是从列表中返回一个字符串，就是为这个会话等待
	// 一个新的消息进入队列
	synchronized String get() {
		String retval;
		while (continueFlag) {
			try {
				retval =   msgs.removeFirst();
				return retval;
			} catch (NoSuchElementException e) {
				try {
					wait();
				} catch (InterruptedException e2) {
				}
			}
		}
		return null;
	}

	public GChat(Socket s, GChatServer m, int me) {
		socket = s;
		myIndex = me;
		mom = m; // 母ChatServer任务
	}

	public void run() {
		// 启动一个线程来读取套接字，然后为所有
		// 聊天对话 把它放到所有的消息队列中
		GChatRead cr = new GChatRead(socket, mom, myIndex);
		cr.start();

		try {
			out = new PrintWriter(socket.getOutputStream(), true);

			String inputLine;

			// 循环拉出队列的消息，然后
			// 把它们写到套接字.
			while ((inputLine = get()) != null) // 从套接字读
			{
				out.println(inputLine);
			}
			socket.close();
		} catch (IOException e) {
			System.out.println("Err: " + e);
		} finally {
			System.out.println("Terminating Chat Thread: " + myIndex);
			try {
				out.close();
				socket.close();

			} catch (Exception e) {
				System.out.println("Exception closing sockets :" + e);
			}
		}
	}// 结束run
}

class GChatRead extends Thread {
	GChatServer mom;
	Socket sock;
	BufferedReader in = null;
	int myIndex;

	GChatRead(Socket s, GChatServer m, int index) {
		sock = s;
		mom = m;
		myIndex = index;
	}

	public void run() {
		// 读取套接字，并为所有会话把字符串读取到所有的消息队列
		try {

			in = new BufferedReader(
					new InputStreamReader(sock.getInputStream()));

			String inputLine;

			while ((inputLine = in.readLine()) != null) // Read from socket
			{
				mom.putMsgs(inputLine);
			}
		} catch (IOException e) {
			System.out.println(" Err: " + e);
		} finally {
			try { 
				in.close();
				sock.close();
				System.out.println(myIndex + ": ChatRead Thread Exitting");
				mom.removeMe(myIndex);
			} catch (Exception e) {
			}
		}
	}// 结束run
}